/*                               -*- Mode: C -*- 
 * mputil.c --- 
 * Author          : Sue Rao
 * Created On      : Mon Jun 23 16:37:08 1997
 */

# include "mputil.h"

tuple ident = NULL;


char *whatTypeDef(list *rhsList)
/**************************************************
 * returns the type of typedef, given the type of instance
 * of rhs term
 */
{
  char *typed;

  typed = NULL;
  if (! emptyList(rhsList)) {
    if(isOfClass(rhsList, _field)){
      typed = (char *)malloc(strlen("RecordTypeDef")+1);
      strcpy(typed, "RecordTypeDef");
    }   
    else if(isOfClass(rhsList, _enumerationconst)){
      typed = (char *)malloc(strlen("EnumeratedTypeDef")+1);
      strcpy(typed, "EnumeratedTypeDef");
    }   
    else {
      typed = (char *)malloc(strlen("TypeDef")+1);
      strcpy(typed, "TypeDef");
    }
  }
  else {
    typed = (char *)malloc(strlen("TypeDef")+1);
    strcpy(typed, "TypeDef");
  }

 return(typed);
}


bool isOfClass(list *rhsList, class_t class)
/**************************************************
 * returns true/false depending on whether the rhs term is an
 * instance of class
 */
{
  listElt *tmp;
  bool found, retn;

  found = false;
  retn = false;
  tmp = rhsList->first;

  
  while ( (tmp != NULL) && (TPTYPE(topElement(rhsList)) != _instance) )
    tmp = tmp->next;

  if (tmp != NULL) {
    if (INSTARG2(topElement(rhsList)) == class)
      retn = true;
    else
      retn = false;
  }
  else
    retn = false;

  return(retn);
}


bool anonymType(list *type)
/*************************************************** 
 * returns true if Type should be anonymous 
 */
{
  if ( (! emptyList(type)) && 
	   (INSTARG2(topElement(type)) == _typeuseexistence) )
    return(FALSE);
  else
    return(TRUE);
}


char *whatName(list *rhsList, int ident)
/**************************************************
 * returns the type's name in the rhs term or if none returns ANONYMOUS
 */
{
  char *name;

  if ( ! anonymType(rhsList))
    return(TPIDENT(topElement(rhsList)));
  else {
    name = (char *)malloc(strlen("ANONYMOUS")+1);
    strcpy(name,"ANONYMOUS");

    name = expandName(name, ident);      
    return(name);
  }
}




char *useNumb(char *name)
/**************************************************
 * add a 'sign' in from of name to be used when
 * the name is used in order to distinguish from 
 * definition
 */
{
  char *new;

  new = (char *)malloc(strlen(name) + 2);
  sprintf(new, "0%s", name);
  return(new);
}

void membersSpecialTypeDef(char *lhsName, list *rhsList)
/**************************************************
 * generates the fact tuples isFieldMemberOf/
 * isEnumerationMemberOf for the rhs term 
 */
{
  char *reln;
  tuple fact;

  reln = NULL;
  if ( ! emptyList(rhsList) ) {
    if (INSTARG2(topElement(rhsList)) == _field)
      reln = "isFieldMemberOf";
    else if (INSTARG2(topElement(rhsList)) == _enumerationconst)
      reln = "isEnumerationMemberOf";
  rhsList = factTupleAttrList(rhsList);
  }
  if (reln != NULL) {
    while (! emptyList(rhsList)) {
      fact = delElement(rhsList);
      factTuple(reln, TPIDENT(fact), lhsName);
    }
  }
}


void makeTypeDef(char *lhsName, list *rhsList, int ident, char *srcName)
/**************************************************
 * generates the fact tuple INSTANCE ... TypeDef/RecordTypeDef/EnumeratedType
 * and definedBy.
 */
{
  factTuple("$INSTANCE", lhsName, whatTypeDef(rhsList));
  factTuple("definedBy", lhsName, srcName);

  if (isOfClass(rhsList, _typeuseexistence))
    rhsDef( whatName(rhsList, ident), srcName);
  else if (! isOfClass(rhsList, _datauseexistence))
    membersSpecialTypeDef( lhsName, rhsList);
}


void lhsDatumDef(char *lhsName, char *rhsName, char *srcName)
/**************************************************
 * generates the fact tuples namely INSTANCE(datumdef)
 * and definedBy, type
 */
{
  factTuple("$INSTANCE", lhsName, "DatumDef");
  factTuple("definedBy", lhsName, srcName);
  factTuple("type", lhsName, useNumb(rhsName));
}


void rhsDef( char *rhsName, char *srcName)
/**************************************************
 * generates the fact tuples namely INSTANCE(typeuseexistence)
 * and usedInSource.
 * Print these tuples in M-Pascal parser. 
 */
{
  factTuple("$INSTANCE", useNumb(rhsName), "TypeUseExistence");
  factTuple("usedInSource", useNumb(rhsName), srcName);
}


list *factTupleAttrList(list *anyList)
/**************************************************
 * puts the items from the given list into the mainlist if
   there are attributes or call factTuple to eventually
   print a relation line if inside M-Pascal parser.
 */
{
  listElt *tmp;
  
  for (tmp=anyList->first; tmp != NULL; tmp = tmp->next) {
    switch (TPTYPE(tmp->item)) {
    case _attribute:
      factAttribute(tmp->item,AttrList_temp);
      break;
    case _instance:
      factTuple("$INSTANCE", INSTARG1(tmp->item), 
				instance2str(INSTARG2(tmp->item)));
      break;
    case _relation:
      factTuple(talRel2str(REL_REL(tmp->item)), RELARG1(tmp->item),
				RELARG2(tmp->item));
      break;
    }
  }

  return(anyList);
}




tuple makeNode(char *name, int start, int end)
/**************************************************
 * generates  a tuple and  assignes identifier, startChar,
 * endChar and returns it
 */
{
  tuple fact;

  fact = initTuple();
  TPIDENT(fact) = name;
  TPTYPE(fact)  = _attribute;
  ATTRARG(fact,_startChar) = itoStr(start);
  ATTRARG(fact,_endChar)   = itoStr(end);

  return(fact);
}


void tupForAnonymousType(char *rhsName, char *srcName,int start, int end)
/**************************************************
 *
 */
{
  list *tmp;
    tmp = initList();
    addElement(tmp, makeNode(rhsName, start, end));
    addElement(tmp, makeTuple("$INSTANCE", useNumb(rhsName), "TypeUseExistence"));
    addElement(tmp, makeTuple("usedInSource", useNumb(rhsName), srcName));
    tmp = factTupleAttrList(tmp);
    delList(tmp);

}


list *varDecl(list *lhsList, list *rhsList, char *srcName, int start, int end)
/**************************************************
 * generates the typedef for the rhs term if rhs is typeuseexistence    
 * or generates the tuples for the lhs datumdef and rhs def
 */
{
  char *rhsName;
  tuple fact;
  listElt *tmp;

  rhsName = whatName(rhsList, start);
  /* note: 'start' is used as a unique identifier for the name */

  if ( emptyList(rhsList) || (! isOfClass( rhsList, _typeuseexistence)) ) {
    makeTypeDef( rhsName, rhsList, start, srcName);
    factTupleAttrList( rhsList);
    /* NA: bug fixing, this was not a loop */
    for (tmp = lhsList->first; tmp != NULL; tmp = tmp->next)
      lhsDatumDef(TPIDENT(tmp->item), rhsName, srcName);
  }
  else
    for (tmp = lhsList->first; tmp != NULL; tmp = tmp->next) {
      lhsDatumDef(TPIDENT(tmp->item), rhsName, srcName);
      rhsDef(rhsName, srcName);
    }
  
  if (strcmp(truncateName(rhsName), "ANONYMOUS") == 0) 
    tupForAnonymousType(rhsName, srcName, start, end) ;

  return(lhsList);
}


list *lhsFormalArgs(list *lhsList, char *srcName)
/**************************************************
 * generates the declaredAsFormalArgsIn tuple and returns 
 * the lhslist
 */
{
  listElt *tmp;
  
  for(tmp = lhsList->first; tmp != NULL; tmp = tmp->next)
    factTuple("declaredAsFormalArgsIn", TPIDENT(tmp->item), srcName); 

  return(lhsList);
}


list *assignAttr(list *lhsList, attribType attr, bool value)
/**************************************************
 * assigns the given attribute to true/false and returns
 * the lhslist 
 */
{
  listElt *tmp;
  
  for(tmp=lhsList->first; tmp != NULL; tmp = tmp->next)
    ATTRARG(tmp->item,attr) = itoStr(value);
  
  return(lhsList);
}

void delList(list *anyList)
/**************************************************
 * deletes the list of tuples
 */
{
  tuple fact;

  while (! emptyList(anyList)) {
    fact = delElement(anyList);
    delTuple(fact);
  }
}


list *concatList(list *l1, list *l2)
/**************************************************
 * concatenates list l2 at the end of list l1
 */
{
  list *l;

  if (emptyList(l1))
    l = l2;
  else {
    l = l1;
    if (l2->last != NULL) {
      l->last->next = l2->first;
      l->last = l2->last;
    }
  }

  return(l);
}


list *makeDUE(list *givenList, char *srcName)
/**************************************************
 * generates a list of tuples namely DUE, usedInSource
 * for the identifiers in the given list
 */
{
  list *newList;
  listElt *tmp;
  
  newList = initList();
  for(tmp = givenList->first; tmp != NULL; tmp = tmp->next) {
    addElement(newList, makeTuple("INSTANCE", TPIDENT(tmp->item),
								  "DataUseExistence"));
    addElement(newList, makeTuple("usedInSource", TPIDENT(tmp->item),
								  srcName)); 
  }

  delList(givenList);
  return(newList);

}


list *makeUsedInSource(list *givenList, char *srcName)
/**************************************************
 * generates a list of tuples namely  usedInSource
 * for the identifiers in the given list
 */
{
  list *newList;
  listElt *tmp;
  
  newList = initList();
  for(tmp= givenList->first; tmp != NULL; tmp = tmp->next)
    if (TPTYPE(tmp->item) == _instance)
      if ( (INSTARG2(tmp->item) == _datauseexistence) ||
		   (INSTARG2(tmp->item) == _typeuseexistence) )
	addElement(newList, makeTuple("usedInSource", INSTARG1(tmp->item),
								  srcName)); 
  delList(givenList);
  return(newList);
 
}


list *makeTypeUseExistence(list *givenList, char *srcName)
/**************************************************
 * generates a list of tuples namely  usedInSource
 * for the identifiers in the given list
 */
{
  list *newList;
  listElt *tmp;
  
  newList = initList();
  for(tmp = givenList->first; tmp != NULL; tmp = tmp->next)
    addElement(newList, makeTuple("$INSTANCE", useNumb(TPIDENT(tmp->item)),
								  "TypeUseExistence")); 
  delList(givenList);
  return(newList);
 
}

list *makeFieldType(list *l1, int ident, list *l2)
/**************************************************
 * Returns a list of tuple "type identifier type_element"
   For each identifier in list l1 with type_element as the 
   type of l2.
 */
{
  list *newList;
  listElt *tmp;

  newList = initList();
  if (! emptyList(l2))
    for (tmp = l1->first; tmp != NULL; tmp = tmp->next)
	  addElement(newList, makeTuple("type", TPIDENT(tmp->item),
									useNumb(whatName(l2, ident))));	

  return(newList);
}


/**************************************************
 * Local Variables: 
 * tab-width: 4
 * End:
 */
 
