/*                               -*- Mode: C -*- 
 * tuple.c --- utilities for the tuples
 * Author          : Sue Rao
 * Created On      : Fri May 16 15:06:13 1997
 */

# include "tuple.h"


attribType str2talAttr( char *attrName)
/**************************************************
 * returns the enumerated type of attribute given the string
 */
{
  if (strcmp(attrName, "dateChanged") == 0)
    return(_dateChanged);
  else if(strcmp(attrName, "endChar") == 0)
    return (_endChar);
  else if(strcmp(attrName, "isConst") == 0)
    return (_isConst);
  else if(strcmp(attrName, "isEntry") == 0)
    return (_isEntry);
  else if(strcmp(attrName, "isFunct") == 0)
    return (_isFunct);
  else if(strcmp(attrName, "path") == 0)
    return (_path);
  else if(strcmp(attrName, "size") == 0)
    return (_size);
  else if(strcmp(attrName, "startChar") == 0)
    return (_startChar);
  else if(strcmp(attrName, "version") == 0)
    return (_version);

}


char *talAttr2str( attribType i)
/**************************************************
 * Returns the attribute "name" given the enumerated attribute type
 */
{
  switch( i) {
  case _dateChanged:  return("dateChanged");
  case _endChar:  return("endChar");
  case _isConst:  return("isConst");
  case _isEntry:  return("isEntry");
  case _isFunct:  return("isFunct");
  case _path:  return("path");
  case _size:  return("size");
  case _startChar:  return("startChar");
  case _version:  return("version");
  }
}


talRel str2TalRel(char *relnTuple)
/**************************************************
 * returns the enumerated type of relation given the string
 */
{ 
  talRel reln;

  if (strcmp(relnTuple, "containingSource") ==  0)
    reln = _containingSource;
  else  if (strcmp(relnTuple, "potentiallyCalledBy") == 0)
    reln = _potentiallyCalledBy;
  else if (strcmp(relnTuple, "foundInSource") == 0)
    reln = _foundInSource;
  else if (strcmp(relnTuple, "definedBy") == 0)
    reln = _definedBy;
  else if (strcmp(relnTuple, "isEnumerationMemberOf") == 0)
    reln = _isEnumerationMemberOf;
  else if (strcmp(relnTuple, "isFieldMemberOf") == 0)
    reln = _isFieldMemberOf;
  else if (strcmp(relnTuple, "declaredAsFormalArgsIn") == 0)
    reln = _declaredAsFormalArgsIn;
  else if (strcmp(relnTuple, "potentiallyIncludedIn") == 0)
    reln = _potentiallyIncludedIn;
  else if (strcmp(relnTuple, "returnType") == 0)
    reln = _returnType;
  else if (strcmp(relnTuple, "successorInCompound") == 0)
    reln = _successorInCompound;
  else if (strcmp(relnTuple, "type") == 0)
    reln = _ofType;
  else if (strcmp(relnTuple, "usedInSource") == 0)
    reln = _usedInSource;
  
  return(reln);

}


class_t str2Instance(char *clss)
/**************************************************
 * returns the enumerated type of instance given the string
 */
{
  return( str2cdbClass(clss));
} 


char *instance2str(class_t clss)
/**************************************************
 * returns the string given the enumerated type of instance
 */
{
  return( cdbClass2str(clss));
}


char *talRel2str(talRel reln)
/**************************************************
 * returns the string given the enumerated type of relation
 */
{
  switch(reln) {
  case _containingSource:       return("containingSource");
  case _potentiallyCalledBy:    return("potentiallyCalledBy");
  case _foundInSource:          return("foundInSource");
  case _definedBy:                return("definedBy");
  case _isEnumerationMemberOf:  return("isEnumerationMemberOf");
  case _isFieldMemberOf:        return("isFieldMemberOf");
  case _declaredAsFormalArgsIn: return("declaredAsFormalArgsIn");
  case _potentiallyIncludedIn:  return("potentiallyIncludedIn");
  case _returnType:             return("returnType");
  case _successorInCompound:    return("successorInCompound");
  case _ofType:                 return("type");
  case _usedInSource:           return("usedInSource");
  }
}


tuple initTuple()
/******************************************************
 *   initialize to NULL 
 */
{
  tuple t;
  int i;
  char* temp;
 
  t = NEW(struct Tuple);
  for (i=0; i<_lastAttr; i++) 
    ATTRARG(t,i) = NULL; 
  t->identifier=NULL;

  return(t);
}


int cmpTuple( tuple t1, tuple t2)
/**************************************************
 * Compares the two tuples and returns 1, 0, -1
 * depending on whether the t1 is >, =, < than the t2
 */
{
  int retn;

  if (strcmp(TPIDENT(t1), TPIDENT(t2)) == 0 ) {
   
    if ( (int)TPTYPE(t1) > (int)TPTYPE(t2))
      retn = 1;
    
    else if( (int)TPTYPE(t1) == (int)TPTYPE(t2)) {
	  if ( (TPTYPE(t1) == _relation) && (REL_REL(t2) != _lastRel) ) {

	  if (REL_REL(t1) == REL_REL(t2))
	    retn = 0;
	  else {

	    if(REL_REL(t1) > REL_REL(t2))
	      retn = 1;
	    else 
	      retn = -1;
	  }    /* if (REL_REL(t1) == REL_REL(t2)) ... else */
	}      /* if (TPTYPE(t1) == relation) && ... */
	else 
	  if ((TPTYPE(t1) == _instance) && (INSTARG2(t2) < 50)) {
	    if (INSTARG2(t1) > INSTARG2(t2))
	      retn = 1;
	    else
	      if(INSTARG2(t1) == INSTARG2(t2))
		retn = 0;
	      else 
		retn = -1;
	  }
	  else
	    retn = 0;
	}        /* if (TPTYPE(t1) == TPTYPE(t2)) */
	else
	  retn = -1;
  }            /* if (strcmp(TPIDENT(t1), TPIDENT(t2)) == 0) */
  else
    retn = strcmp(TPIDENT(t1), TPIDENT(t2));
  
  return(retn);
}


void printFTuple(FILE *fout, tuple element)
/**************************************************
 * Prints a single tuple in a an output file
 */
{
  int i;

  switch(TPTYPE(element)) {
  case _attribute: 
    fprintf(fout, "%s { ", ATTRARG1(element));
    for(i = 0; i< _lastAttr; i++) 
      if( ATTRARG(element,i) != NULL)
		fprintf(fout,"\n\t%s = %s", talAttr2str(i), ATTRARG(element,i));
    
    fprintf(fout,"}\n");
    break;

  case _instance:
    fprintf(fout,"$INSTANCE %s %s\n", INSTARG1(element),
			                          instance2str(INSTARG2(element)));
    break;

  case _relation:
    fprintf(fout,"%s %s %s\n", talRel2str(REL_REL(element)),
		                       RELARG1(element),
		                       RELARG2(element));
    break;  

  default:
    break;
  }
}

void printTuple(tuple element)
/**************************************************
 * Prints a single tuple to default output
 */
{
  printFTuple(stdout,element);
}

void delTuple(tuple element)
/**************************************************
 * deletes a single tuple element
 */
{
  int i;
  
  if (element != NULL) {
	if (TPIDENT(element) != NULL) {
	  free(TPIDENT(element));
	  TPIDENT(element) = NULL;
	}

    switch(TPTYPE(element)) { 
    case _attribute:  
      for(i = 0; i< _lastAttr; i++)  
        if ( ATTRARG(element,i) != NULL) {
		  free(ATTRARG(element,i));
		  ATTRARG(element,i) = NULL;
		}
	  break; 
	case _instance: 
	  break; 
	case _relation: 
	  if (RELARG2(element) != NULL) {
		free(RELARG2(element)); 
		RELARG2(element) = NULL;
	  }
	  break;  
	} 

	free(element);  
  }
}


tuple copyTuple(tuple element)
/**************************************************
 * returns a copy of element
 */
{
  int i;
  tuple copy;

  if (element != NULL) {
	copy = initTuple();
	if (TPIDENT(element) != NULL)
      TPIDENT(copy) = (char *)dupString(TPIDENT(element));

	TPTYPE(copy) = TPTYPE(element);
    switch(TPTYPE(element)) { 
   
	case _attribute:  
      for(i = 0; i < _lastAttr; i++) 
        if ( ATTRARG(element,i) != NULL) 
		  ATTRARG(copy,i) = (char *)dupString(ATTRARG(element,i));
	  
	  break;

	case _instance: 
	  INSTARG2(copy) = INSTARG2(element);
	  break;

	case _relation:
	  REL_REL(copy) = REL_REL(element);
	  if (RELARG2(element) != NULL) 
		RELARG2(copy) = (char *)dupString(RELARG2(element));
	  break;   
	} 
  }
  return(copy);
}


tuple insertTuple(tuple list, tuple newOne)
/**************************************************
 * creates the tuple and stores it in the list, 
 * returns the new list. The list should be sorted on 
 * identifier's name.
 */
{
  int cmp;
  type reln;
  tuple previous, current, following;
  
  if (list == NULL)
    list = newOne;
  else {
    following = list;
    current   = NULL;
    previous  = NULL;
    cmp =  0;

    while ( (cmp <= 0) && (following != NULL) ) {
      previous  = current;
      current   = following;
      following = TPNEXT(following);

      cmp = cmpTuple( current, newOne);
    }
    
    if ( cmp > 0) {
      TPNEXT(newOne) = current;
      if (previous != NULL)
		TPNEXT(previous) = newOne;
      else
		list = newOne;
    }
    else
      TPNEXT(current) = newOne;
  }
  
  return( list);
}


tuple searchTuple(tuple list, type reln, char *name) 
/**************************************************
 * Search for the Tuple in the list
 * if relation code is -1 then returns first tuple with the name
 */
{
  tuple  searched;
  bool found, exist;
  int cmp;

  found = false;
  exist = true;

  searched = initTuple();

  TPTYPE(searched)  = reln;
  TPIDENT(searched) = name;
  if (reln == _relation)
    REL_REL(searched) = _lastRel;
  else 
    if (reln == _instance)
      INSTARG2(searched) = 50;

  while ( (! found) && (list != NULL) && (exist) ) {
    cmp = cmpTuple( list, searched);
    if ( cmp == 0)
      found = true;
    else 
      if (cmp > 0)
	exist = false;
      else 
	list = TPNEXT(list);
  }

  free(searched);  /* don't use delTuple because `searched' borrows
                      all its memory (strings) from other places */

  if (found)
    return( list);
  else
    return( NULL);

}


int cmpRelation(tuple t1, tuple t2)
/**************************************************
 * compares two tuples t1, t2 on identifier, tupleType and relName
 * and returns either 1, 0, -1. 
 */
{
  int retn;

  if(strcmp(TPIDENT(t1), TPIDENT(t2)) == 0) 
    if ( (int)TPTYPE(t1)  == (int)TPTYPE(t2) )
      if ( (int)REL_REL(t1) == (int)REL_REL(t2) )
		retn = 0;
      else if ( (int)REL_REL(t1) > (int)REL_REL(t2) )
		retn = 1;
	  else
		retn = -1;
    else if ( (int)TPTYPE(t1) > (int)TPTYPE(t2) )
	  retn = 1;
	else
	  retn = -1;
  else
    return(strcmp(TPIDENT(t1), TPIDENT(t2)));
  
  return (retn);
}


tuple searchRelation(tuple list, talRel reln, char * name)
/**************************************************
 * searches for the RELATION tuples and returns the element if found
 * else returns NULL
 */
{
  tuple searched;
  bool found, exist;
  int cmp;

  found = false;
  exist = true;

  searched = initTuple();

  TPTYPE(searched)  = _relation;
  RELARG1(searched) = name;
  REL_REL(searched) = reln;

  while ( (! found) && (list != NULL) && (exist) ) {

    cmp = cmpRelation(list, searched);
    if ( cmp ==0 )
      found = true;
    else 
      list = TPNEXT(list);
  }

  free(searched);  /* don't use delTuple because `searched' borrows
                      all its memory (strings) from other places */


  if (found)
    return( list);
  else
    return(NULL);
}


tuple findPreviousTuple(tuple list, tuple current)
/**************************************************
 * Search for the previous Tuple before current
 * Returns NULL if current not found or first element of list.
 */
{
  bool found;

  found = false;
  
  if (list != current)
    while ( (! found) && (list != NULL) )
      if (TPNEXT(list) == current)
	found = true;
      else
	list=TPNEXT(list);

  if (found)
    return( list);
  else
    return( NULL);

}


tuple removeTuple(tuple list, tuple del)
/**************************************************
 * Removes the tuple from the list
 */
{
  tuple tmp;

  if((list !=NULL) && (del != NULL)) {
    
    if (list == del) {
      list = TPNEXT(del);
      delTuple(del);
    }
    
    else {
      tmp = findPreviousTuple(list, del); 

      if (tmp != NULL) {
		TPNEXT(tmp) = TPNEXT(del);
		delTuple(del);
      }
    }
  }

  return(list);
}


void printTupleList(tuple list)
/**************************************************
 * Print the list of identifiers for which the necessary
 * tuples were not found
 */
{
  while (list != NULL) {
    
    printTuple( list);
    list = removeTuple(list, list);
  }
}

/**************************************************
 * Local Variables: 
 * tab-width: 4
 * End:
 */
