/*                               -*- Mode: C -*- 
 * tuple.h --- utilities for the tuples 
 * Author          : Sue Rao
 * Created On      : Fri May 16 09:47:02 1997
 */
# ifndef TUPLE_H
# define TUPLE_H

# include<stdio.h>
# include<string.h>
# include "c++_c.h"


/* enumerated types to put in the field of the `tuples' (see below ...) */
typedef enum talRel_t {
  _containingSource, _declaredAsFormalArgsIn, _definedBy, _foundInSource,
  _isEnumerationMemberOf, _isFieldMemberOf, _ofType, _potentiallyCalledBy,
  _potentiallyIncludedIn, _returnType, _successorInCompound, _usedInSource,
  _lastRel } talRel;

typedef enum attribType_t {
  _dateChanged, _endChar, _isConst, _isEntry, _isFunct, _path, _size,
  _startChar, _version, _lastAttr } attribType; 

typedef enum type_t { _instance, _attribute, _relation } type; 

/* some macros to simplify the use of the nested struct */
# define TPIDENT(T)   ((T)->identifier)
# define TPTYPE(T)    ((T)->tupleType)
# define TPNEXT(T)    ((T)->next)

# define INSTARG1(T)  (TPIDENT(T))
# define INSTARG2(T)  ((T)->secondArg.secondName)

# define ATTRARG1(T)  (TPIDENT(T))
# define ATTRARG(T,A) ((T)->secondArg.attrValue[A])

# define RELARG1(T)   (TPIDENT(T))
# define REL_REL(T)   ((T)->secondArg.relnTuple.relName)
# define RELARG2(T)   ((T)->secondArg.relnTuple.secondName)


/* the structs needed to define tuples */
typedef struct RelField {
  char *secondName;
  talRel relName;
} relField;

typedef union Second {
  class_t secondName; /* second identifier of  INSTANCE tuple */
  relField relnTuple;   /* struct of reln & 2nd identifier of RELATION tuple */
  char *attrValue[_lastAttr]; /*array of attribute values of ATTRIBUTE tuple */
} second;

typedef struct Tuple {
  type   tupleType;
  char  *identifier;
  second secondArg;
  struct Tuple *next;  /* to be removed */
} *tuple;


/* convertion between enumerated types and strings */
attribType str2talAttr(char*);
char      *talAttr2str(attribType);
talRel     str2TalRel(char*);
class_t    str2Instance(char*);
char      *instance2str(class_t);
char      *talRel2str(talRel);

/* tuple utilities */
tuple initTuple();
int   cmpTuple(tuple,tuple);
int   cmpRelation(tuple,tuple);
void  printTuple(tuple);
void  delTuple(tuple);
tuple copyTuple(tuple element);

/* the following should be moved to tuple-list, which uses a different
   kind of list (should remove the `next' attribute in `tuple' too). */
tuple insertTuple(tuple list,tuple);
tuple searchTuple(tuple list,type, char*) ;
tuple searchRelation(tuple list,talRel,char*);
tuple removeTuple(tuple list,tuple);
void printTupleList(tuple list);
void printFTuple(FILE *fout, tuple element);
# endif
